#include <iostream>
#include <fstream>
#include <sstream>
#include <string>
#include <vector>
#include <algorithm>
#include <ctime>
#include <cstdlib>
#include <cstdio>
#include <cmath>

#include "EPADLocal.h"

EPADLocalSEQUENCE::EPADLocalSEQUENCE(int length_seq, EPADLocal* pModel)
{
	Init(length_seq, pModel);
}

void EPADLocalSEQUENCE::Init(int length_seq, EPADLocal* pModel)
{
	cerr << "EPADLocalSEQUENCE::Init(" << length_seq << ")" << endl;
	forward=NULL;
	backward=NULL;
	m_pModel = pModel;
	this->length_seq = length_seq;
	arrVi=NULL;
	obs_label = new int[length_seq];
	int df = m_pModel->dim_features*length_seq;
	
	_features = new Score[df];

	predicted_label = new int[length_seq];	
	obs_feature = new Score*[length_seq];
	for(int i=0;i<length_seq;i++)
		obs_feature[i] = new Score[m_pModel->dim_one_pos];
}


EPADLocalSEQUENCE::~EPADLocalSEQUENCE(){
	delete forward;
	delete backward;
	delete arrVi;
	delete obs_label;
	delete _features;
	for(int i=0;i<length_seq;i++)
		delete obs_feature[i];
	delete obs_feature;
}

void EPADLocalSEQUENCE::ComputeForward()
{
	if (forward==NULL)
		forward = new ScoreMatrix(m_pModel->num_states,length_seq, "forward");
	forward->Fill(LogScore_ZERO);
	for(int l=0;l<(int)m_pModel->sparseMi1.nnz();l++){
		int i = m_pModel->A1[l];
		int leftLabel=i/m_pModel->num_labels-1;
		if (leftLabel>-1)
			continue;
		(*forward)(i,0)=ComputeScore(DUMMY,i,0);
	}

	for(int t=1;t<length_seq;t++){
		for(int l=0; l< (int)m_pModel->d_sparseMi.nnz(); l++){
			int leftState=m_pModel->d_I[l]; 
			int currState = m_pModel->d_J[l];
			int leftLabel=currState/m_pModel->num_labels-1;
			if (leftLabel==-1)
				continue;
			Score new_score = ComputeScore(leftState,currState,t);
			LogScore_PLUS_EQUALS((*forward)(currState,t),new_score + (*forward)(leftState,t-1));
		}
	}
}

void EPADLocalSEQUENCE::ComputeBackward()
{
	if (backward==NULL)
		backward = new ScoreMatrix(m_pModel->num_states,length_seq, "backward");
	backward->Fill(LogScore_ZERO);
	for(int l=0;l<(int)m_pModel->sparseMi1.nnz();l++){
		int i = m_pModel->A1[l];
		int leftLabel=i/m_pModel->num_labels-1;
		if (leftLabel==-1)
			continue;
		(*backward)(i,length_seq-1)=0;
	}
	for(int t=length_seq-2;t>=0;t--){
		for(int l=0; l< (int)m_pModel->d_sparseMi.nnz(); l++){
			int currState=m_pModel->d_I[l];
			int leftLabel=currState/m_pModel->num_labels-1;
			if ((t>0 && leftLabel==-1) || (t==0 && leftLabel>-1))
				continue;
			int rightState = m_pModel->d_J[l];
			Score new_score = ComputeScore(currState,rightState,t+1);
			LogScore_PLUS_EQUALS((*backward)(currState,t),new_score + (*backward)(rightState,t+1));
		}
	}
}

Score EPADLocalSEQUENCE::GetGateOutput(int gate, Score* features){
	Score output = 0;
	for(int j=0;j<m_pModel->dim_features;j++)
		output+=m_pModel->GetWeightG(gate,j)*features[j]; //dim_features
	return Gate(output);
}

void EPADLocalSEQUENCE::ComputeGates()
{
	gates.resize(m_pModel->num_gates, length_seq);
	for (int pos=0;pos<length_seq; pos++) {
		Score* features=getFeatures(pos);
		for (int k=0;k<m_pModel->num_gates;k++)
			gates(k,pos) = GetGateOutput(k,features);
	}
}

void EPADLocalSEQUENCE::ComputeVi4Sampling()
{
	if (arrVi==NULL)
		arrVi = new ScoreMatrix();
	int num_gates = m_pModel->num_gates;
	int num_labels = m_pModel->num_labels;
	int num_states_o1 = num_labels+1; // for 1-order
	int cnt_m1 = m_pModel->sparseMi1.nnz();
	arrVi->resize(num_states_o1, length_seq);
	for (int pos=0;pos<length_seq; pos++)
		for (int l=0; l<cnt_m1; l++){
			int currState = m_pModel->A1[l];

			int weightLStart = m_pModel->iGateStart + currState*num_gates;
			Score score = 0;
			for(int k=0;k<m_pModel->num_gates;k++) {
				Score output = gates(k,pos);
				score += m_pModel->weights[weightLStart++]*output;
			}
			arrVi->set(currState, pos, score);
		}
}

Score EPADLocalSEQUENCE::ComputeScore(int leftState, int currState, int pos){
	//compute num_gates neurons' output
	int currLabel = currState;
	double mi=m_pModel->GetWeightT(leftState,currLabel);
	double vi = (*arrVi)(currState, pos);
	Score score = vi+mi;
	//if (pos<=1)cerr << pos << ":" << leftState << ":" << currLabel << ":" << vi << ":" << mi << endl;
	return score;
}

void EPADLocalSEQUENCE::makeFeatures(){
	//build features for local windows	
	for(int t=0;t<length_seq;t++){
		int pivot = t*m_pModel->dim_features;
		for(int i=0; i< m_pModel->window_size; i++){
			int offset = t+i-m_pModel->window_size/2;
			if(offset <0 || offset >=length_seq){
				for(int j=0;j<m_pModel->dim_one_pos;j++)
					_features[pivot] = 0, pivot++;
			} else {
				for(int j=0;j<m_pModel->dim_one_pos;j++)
					_features[pivot] = obs_feature[offset][j], pivot++;
			}
		}
		_features[pivot] = 1; // last one for the bias term
	}
}

void EPADLocalSEQUENCE::CalcPartition(){
	Partition = (Score)LogScore_ZERO;
	//Score FP = (Score)LogScore_ZERO;
	Score BP = (Score)LogScore_ZERO;
	for(int k=0;k<m_pModel->num_states;k++)
		LogScore_PLUS_EQUALS(Partition, (*forward)(k,length_seq-1));
	for(int k=0;k<m_pModel->num_states;k++) 
		LogScore_PLUS_EQUALS(BP, (*backward)(k,0) + (*forward)(k,0));
	//cerr << "Partition="<< Partition << "; BP=" << BP << endl;
}


Score* EPADLocalSEQUENCE::getFeatures(int pos){
	int offset = pos*m_pModel->dim_features;
	return _features+offset;
}

void EPADLocal::Initialize(string strModelFile, string strLabelFile, string strDataFile)
{
	double ref[]={0.0020,0.0096,0.0416,0.0046,0.1023,0.0831,0.0019,0.0045,0.0107,0.0079,0.0036,0.0040,0.0100,0.0382,0.0037,0.0082,0.0107,0.0093,0.0042,0.0106,0.0103,0.0050,0.0250,0.0070,0.0066,0.0189,0.0086,0.0051,0.0063,0.0058,0.0089,0.0107,0.0095,0.0118,0.0051,0.0074,0.0091,0.0055,0.0039,0.0046,0.0117,0.0028,0.0047,0.0036,0.0037,0.0084,0.0093,0.0209,0.0052,0.0051,0.0031,0.0092,0.0076,0.0185,0.0067,0.0028,0.0075,0.0028,0.0080,0.0052,0.0133,0.0038,0.0034,0.0101,0.0071,0.0054,0.0062,0.0078,0.0013,0.0038,0.0067,0.0049,0.0053,0.0039,0.0074,0.0066,0.0123,0.0104,0.0035,0.0147,0.0079,0.0043,0.0038,0.0108,0.0030,0.0047,0.0105,0.0044,0.0298,0.0066,0.0055,0.0100,0.0086,0.0012,0.0045,0.0133,0.0144,0.0062,0.0021,0.0180};
	referenceState=new double[100];
	memcpy(referenceState, ref, sizeof(double)*100);
	for (int i=0;i<100; i++)
		cerr<< referenceState[i] << " ";
	cerr << endl;

	int data;
	char buf[10240];
	string tmp;
	cerr << "   Model file("<< strModelFile << endl;
	ifstream fin(strModelFile.c_str()); // no file header
	for (int i=0; i<8; i++) {
		if (!fin.getline(buf,10240))
			break;
		else {
			istringstream si(buf);
			si >> tmp >> data;
			if (tmp=="num_params:")
				num_params = data;
			else if (tmp=="num_gates:")
				num_gates = data;
			else if (tmp=="num_gates2:")
				num_gates2 = data;
			else if (tmp=="window_size:")
				window_size = data;
			else if (tmp=="dim_features:")
				dim_features = data;
			else if (tmp=="num_labels:")
				num_labels = data;
			else if (tmp=="dim_one_pos:")
				dim_one_pos = data;
			else if (tmp==" weights:" || tmp.find("weights:")!=string::npos)
				break;
			cerr << tmp << "::::"<<data<<endl;
		}
	}

	int nRows;
	num_states=num_labels;
	iGateStart = num_labels*(num_states+1);
	iFeatureStart = iGateStart+num_gates*num_states;
	nRows=iGateStart/2;
	cerr << "1order CNFSampler with num_states=" << num_states << "; iGateStart=" << iGateStart << endl;

	bool* mtrMi1= new bool[num_states];
	memset(mtrMi1, 0, sizeof(bool)*num_states);
	sparseMi1=CCoorMatrix<int>(num_states, 1, nRows);
	sparseMi1.empty();
	cerr << num_states*num_labels/1024<<":::"<<sizeof(bool)*num_states*num_labels/1024<<"K"<<sizeof(bool)<<endl;
	bool* mtrMi= new bool[num_states*num_labels];
	memset(mtrMi, 0, sizeof(bool)*num_states*num_labels);

	if (weights)
		delete weights;
	weights= new double[num_params];
	int mi_cnt = 0;
	for(int i=0;i<num_params;i++){
		fin >> weights[i];
		if (weights[i]!=0) {
			//cerr << weights[i] << " ";
			mi_cnt++;
			if (i>=iGateStart) continue;

			int s1=i/num_labels-1;
			int c=i%num_labels;
			int s2=c; // for 1-order
		
			if (mtrMi1[s2]==0){
				mtrMi1[s2]=1;
				sparseMi1.insert(s2,1)=s2;
			}
			if (s1>=0 && mtrMi1[s1]==0){
				mtrMi1[s1]=1;
				sparseMi1.insert(s1,1)=s1;
			}
		}
	}
	//cerr << "mi_cnt=" << mi_cnt<< endl;
	fin.close();

	ifstream flabel(strLabelFile.c_str());
	int num_label=0;
	flabel >> tmp >> num_label; // the first line of the label file should be: num_label: 102
	cerr <<num_states<< "label file(" << strLabelFile << "): " << tmp << num_label << endl;
	if (tmp=="num_label:") {
		int l;
		string s;
		labelMapI2S.resize(num_label);
		for(int i=0;i<num_label;i++) {
			flabel >> l >> s; //cerr << i << " " << l << ":" << s << endl;
			addLabel(l,s);
		}
	}

	delete mtrMi1;
	flabel.close();
	
	sparseMi1.internal_order();
	A1= sparseMi1.getA();

	d_sparseMi=CCoorMatrix<double>(num_states+1, num_states, mi_cnt+1);
	d_sparseMi.empty();
	cerr << " Compute Transition Matrix "<<mi_cnt<<endl;

	mmap=new int[num_states];
	memset(mmap, 0, sizeof(int)*num_states);

	Mi.resize(num_states+1, num_labels);
	for (int l=0;l<(int)sparseMi1.nnz();l++){ // number of valid states
		int i=A1[l];
		mmap[i]=l;
		for (int j=0; j<num_labels; j++){
			Mi(i+1,j) = exp(GetWeightT(i,j));;
		}
	}
	delete mtrMi;

	d_sparseMi.internal_order();

	cerr <<"iGateStart:="<< iGateStart << " d_sparseMi:=" << d_sparseMi.nnz() << endl;
	d_I= d_sparseMi.getI(); //I is the row index array
	d_J= d_sparseMi.getJ(); //J is the column index array
	d_A= d_sparseMi.getA(); // A points to the content of sparseMi
	crow = d_sparseMi;
		rowR= crow.getR();
		rowJ= crow.getJ();
		rowA= crow.getA();
  	ccol = d_sparseMi;
		colI= ccol.getI();
		colC= ccol.getC();
		colA= ccol.getA();

	LoadSamplerData(strDataFile);
	ComputeVi4Sampling();
	residue_energy.resize(seq_length, num_labels);
}

double EPADLocal::Calculate1orderEnergy(vector<int >& labels, int start, int end)
{
	double energy = 0.0;
	int seq_len= (int)labels.size();
	int num_labels = num_states;

	//calculate the sampling possibilities at every postition for each label, and store in memory(i, j)
	for (int i = start; i <= end; i++) {
		// for all the possible labels at the position "i"
		for (int j = 0; j < num_labels; j++) {
			double transmission = 0.0; 
			memory.set(i,j, 0.0);
			//loop through all the possible labels at position i-1
			if (i==0)
				transmission = 1;
			else if (i==start){
				transmission = Mi(labels[start-1]+1,j);
				//fprintf(stderr, "**%d:%4.3f:%d** ", labels[start-1], transmission, j);
			} else {
				for (int k = 0; k < num_labels; k++) 
					transmission += memory(i-1, k) * Mi(k+1,j);
			}
			memory.set(i,j, transmission * arrVi(i,j));
			//fprintf(stderr, "%4.3f=%4.2f*%4.2f ", memory(i,j), transmission, arrVi(i,j));
		} //cerr << endl;
		scale[i] = memory.row_normalize(i);
	}

	int h=-1; // h is the state at position i+1. 
	if (end<seq_len-1) 
		h = labels[end+1]; //if end<seq_len-1, the state for positon end+1 is from the input sequence
 	double* p=new double[num_labels];
	memset(p, 0, sizeof(double)*num_labels);
	for (int i = end; i >= start; i--) {
		double scale=0.0;
		//loop through all the possible states for position i; 
		for (int j = 0; j < num_labels; j++) {
			if (i==seq_len-1)
				p[j] = memory(i, j);
			else //if (i>=seq_len-2 || h==-1)
				p[j] = memory(i,j) * Mi(j+1,h);
			scale+=p[j];
		} //cerr << endl;
		h=labels[i];
		double res_energy=-log(p[h]/scale/referenceState[h]);
		energy+=res_energy;
	}
	delete p;
	return energy;
}

double EPADLocal::CaculateResidueEnergy(int idx, int label)
{
	double energy=residue_energy(idx, label);
	//if (energy==0){
		residue_energy(idx, label)=-log(arrVi(idx, label)/referenceState[label]);
		energy=residue_energy(idx, label);
	//}
	fprintf(stderr, "%5.1f", energy);
	return energy;
}

void EPADLocal::ComputeVi4Sampling()
{
	cerr << " Compute State Matrix -- "<<endl;
	seq4sampling.ComputeGates();

	int cnt_m1 = sparseMi1.nnz();
	cerr << " ComputeVi with "<<cnt_m1 << " states..."<< endl;
	seq4sampling.ComputeVi4Sampling();

	arrVi.resize(seq_length, num_states);
	for (int i=0; i<seq_length; i++){
		for (int l=0; l<cnt_m1; l++){
			int j = A1[l]; 
			Score s = (*seq4sampling.arrVi)(j,i);
			arrVi.set(i,j, exp(s));
		}
	}

	memory.resize(seq_length, num_states);
	m=new int[num_states]; m0=new int[num_states]; mm=new int[num_states];
	labels=new int[seq_length]; states=new int[seq_length];
	scale= new double[seq_length];
}

void EPADLocal::LoadSamplerData(string strDataFile)
{
	ifstream seq_in(strDataFile.c_str());
	num_data = 1; // num_data should be equal to 1
	// construct a new sequence
	seq_in >> seq_length;
	string stmp;
	getline(seq_in, stmp);
	seq4sampling.Init(seq_length, this);

	cerr << "LoadSamplerData(" << strDataFile << "): " << seq_length << endl;
	for(int j=0;j<seq_length;j++){
		for(int k=0;k<dim_one_pos-1;k++){
			seq_in >> seq4sampling.obs_feature[j][k];
		}
		seq4sampling.obs_feature[j][dim_one_pos-1] = 1;
	}
	seq_in.close();
	seq4sampling.makeFeatures();
}
