//this file defines classes for all the scoring functions

#ifndef SCORINGFUNCTION
#define SCORINGFUNCTION

#include <vector>

#include "residue.h"
#include "ScoreMatrix.h"
#include "EPAD.h"
#include "EPADLocal.h"

using namespace std;

//ScoringFunction is the base class of all the other scoring functions
class ScoringFunction
{
public:
	ScoringFunction();
	virtual ~ScoringFunction();

	virtual int Init(string, string, string="");

	//calculate the score of a given structure
	virtual double evaluate(vector<RESIDUE >& structure);
	virtual double evaluate(vector<RESIDUE >& structure, int start, int end);

	//more evaluate functions to be added here for different representation of a structure
	//calculate the radius of a given protein
	double CalculateRadius(vector<RESIDUE>& structure);

public:
	string mName; // This is the name of the energy
	int startRes, endRes, indicator;
	double NEFF;
	int model_typ;

	vector<int> residue_flag; // 0--residue is missed; 1-- residue is real. We won't predict contact for those missing residues
	vector<double> pointEnergies;
};

typedef vector<RESIDUE > BACKBONE;

// This class provides interfaces to calculate the short range (chain distance < 6) EPAD potential 
// of a protein structure. 
class ScoringByEPADLocal: public ScoringFunction
{
public:
	ScoringByEPADLocal();
	~ScoringByEPADLocal();

	/* 
	This function takes as input the model file modelCNFo1SS.dat in the config directory, 
	the *.epad_local file, and another auxiliary label file cnf_label_SS_1.dat in the config 
	directory. It initializes the EPADLocal potential function and generates the distance 
	distribution of local residue pairs. It returns 1 if initialization is successful and 0 
	otherwise.

	Example: 
		ScoringFunction* sf=new ScoringByEPADLocal();
		sf->Init("./config/modelCNFo1SS.dat", "./SEQ/EPAD/1ektA. epad_local", "./config/cnf_label_SS_1.dat");
	*/
	int Init(string modelFile, string feature_file, string labelFile="");

	//calculate the score of a given structure
	double evaluate(vector<RESIDUE >& structure);

	/*
	This function takes a peptide structure as input, and returns the EPADLocal potential 
	between the residue start and residue end. Residues are numbered starting from 0. If the 
	input parameter end is -1, it indicates the last residue.

	Example:
		vector<RESIDUE> g_backbone; // initialized somewhere before, containing the 3D structure of a decoy
		double energy = sf ->evaluate(g_backbone, 0, -1);
	*/
	double evaluate(vector<RESIDUE >& structure, int start, int end);

protected:
	EPADLocal model;
	ScoreMatrix res_energy;

	int fragLen; // fragment length
	int libLen; // library size
	string fragLibFile;
	double ***fragments;

	void readLib();
	void ReLabel(int start, int end, BACKBONE& protein, vector<int >& labels);
};

// This class provides interfaces to calculate the long range (chain distance >= 6) EPAD 
// potential of a protein model. 
class ScoringByEPAD: public ScoringFunction
{
public:
	ScoringByEPAD();
	~ScoringByEPAD();

	// This function reads in the configuration file epad.cfg and the *.epad file, initializes 
	// the potential function and generates the distance distribution of each residue pair. It 
	// returns 1 if initialization is successful and 0 otherwise.
	int Init(string config_file, string feature_file, string dummy="");

	//calculate the score of a given structure
	double evaluate(vector<RESIDUE >& structure);

	// This function takes a protein structure as input, and returns the EPAD potential 
	// between the residue start and residue end. The residue number starts from 0. If 
	// the input parameter end is -1, it indicates the last residue.
	double evaluate(vector<RESIDUE >& structure, int start, int end);

	enum {FUNC_CONTACT_ENERGY_FM=3, FUNC_CONTACT_ENERGY_TB=4, FUNC_CONTACT_PAIRWISE_ENERGY=9, FUNC_CONTACT_DISTRIBUTION_ENERGY=10, 
				 FUNC_CONTACT_DISTRIBUTION=11, FUNC_CONTACT_PREDICTED_ENERGIES=12, FUNC_CONTACT_CB_DISTRIBUTION=13} mfunc;

protected:
	EPAD model;
	ScoreMatrix res_energy;
	bool predicted;
};

#endif
