#include <iostream>
#include <fstream>
#include <sstream>
#include <string>
#include <vector>
#include <algorithm>
#include <ctime>
#include <cstdlib>
#include <cstdio>
#include <cmath>
#include "constraints.h"

Constraints::Constraints()
{
}

// get the secondary structure index i of a residue: i=-100 for err; i=-1 for loop; i=0->99 denotes the i-th beta strand; i=100 and up denotes the (i-100)-th alpha helix
int Constraints::getSSIdx(int residue) 
{
	if (residue<0 || residue>=seq_len)
		return -100;
	return ss_idx[residue];
}

void Constraints::getSecondaryStructure(int seq_len, const int* ss)
{
	char* sn="CHB";
	int last_beta=-1, last_alpha=99; // for alpha, 99=100-1 denotes -1
	beta_idx=-1; alpha_idx=99; 
	this->seq_len=seq_len;
	ss_idx.resize(seq_len, -1); // for each residue, we assign it i if it's on the i-th strand, and 100+i if it's on the i-th helix, -1 for loop
	for (int i=0; i<seq_len; i++){
		switch (ss[i])
		{
		case 2: // beta strand
			if (last_beta==-1){
				beta_idx++;
				BetaStruct betaS;
				betaS.start = i;
				betas.push_back(betaS);
				if (i>0 && ss[i-1]==1 && alpha_idx>=100)
					alphas[alpha_idx-100].end=i-1;
				last_beta=beta_idx;
			}
			ss_idx[i]=beta_idx;
			last_alpha=100-1;
			break;
		case 1: // alpha helix
			if (last_alpha==100-1){
				alpha_idx++;
				AlphaStruct alphaS;
				alphaS.start = i;
				alphas.push_back(alphaS);
				if (i>0 && ss[i-1]==2 && beta_idx>=0)
					betas[beta_idx].end=i-1;
				last_alpha=alpha_idx;
			}
			ss_idx[i]=alpha_idx;
			last_beta=-1;
			break;
		default: // coil
			last_alpha=100-1; last_beta=-1;
			if (i>0 && ss[i-1]==2 && beta_idx>=0)
				betas[beta_idx].end=i-1;
			if (i>0 && ss[i-1]==1 && alpha_idx>=100)
				alphas[alpha_idx-100].end=i-1;
			break;
		}
		cerr << ss_idx[i]<<sn[ss[i]]<<":";
	}cerr << endl;
}

void Constraints::generateConstraints(int seq_len, const int* ss, string name)
{
	cerr << "Constraints--generateConstraints" << endl;
	p_name= name;
	int max_beta = seq_len/10, max_alpha= seq_len/10; // max number of beta strand/helix
	getSecondaryStructure(seq_len, ss);
	// d(i,j) plus assistant var indicating contact between beta strands/helix, plus the constant for sum(w_i*x_i)<c_i
	beta_start = seq_len*(seq_len-1)/2;
	alpha_start = beta_start + max_beta*(max_beta-1)/2;
	totalVar = alpha_start + max_alpha*(max_alpha-1)/2 +1;

	// triangle inequalities
	for (int i=2; i<seq_len-4; i++)
	for (int j=i+1; j<seq_len-3; j++)
	for (int k=j+1; k<seq_len-2; k++){
		if (ss_idx[i-2]==-1 && ss_idx[i-1]==-1 &&ss_idx[i]==-1 &&ss_idx[i+1]==-1 &&ss_idx[i+2]==-1
		 	&& ss_idx[j-2]==-1 && ss_idx[j-1]==-1 &&ss_idx[j]==-1 &&ss_idx[j+1]==-1 &&ss_idx[j+2]==-1
				&& ss_idx[k-2]==-1 && ss_idx[k-1]==-1 &&ss_idx[k]==-1 &&ss_idx[k+1]==-1 &&ss_idx[k+2]==-1)
			continue;
		
		Constraint coef; //coef.resize(totalVar, 0);
		coef.setDist(1, i, j, seq_len); 
		coef.setDist(-1, j, k, seq_len); 
		coef.setDist(-1, i, k, seq_len);
		coef.setConstB(totalVar-1,0);
		constraintMat.push_back(coef);

		Constraint coef1; //coef1.resize(totalVar, 0);
		coef1.setDist(1, j, k, seq_len);
		coef1.setDist(-1, i, k, seq_len);
		coef1.setDist(-1, i, j, seq_len);
		coef1.setConstB(totalVar-1,0);
		constraintMat.push_back(coef1);
		
		Constraint coef2; //coef2.resize(totalVar, 0);
		coef2.setDist(1, i, k, seq_len); 
		coef2.setDist(-1, j, k, seq_len); 
		coef2.setDist(-1, i, j, seq_len);
		coef2.setConstB(totalVar-1,0);
		constraintMat.push_back(coef2);
	}
	cerr << seq_len << " " << totalVar << "-" << beta_start << " " << constraintMat.size() << endl;

	// minimum number of required contacts: min(len_r,len_s)-1 for beta strands
	for (int r=0; r<beta_idx-1; r++)
	for (int s=r+1; s<beta_idx; s++){
		BetaStruct betaSr = betas[r], betaSs=betas[s];
		for (int i=betaSr.start; i<=betaSr.end; i++)
		for (int j=betaSs.start; j<=betaSs.end; j++){
			Constraint coef; //coef.resize(totalVar, 0);
			coef.setDist(-1/8, i, j, seq_len); // idx(i,j)=(j-i)+sum(l-t-1), 0<=t<=i-1 ==> idx=il-(3+i)*i/2)+j
			coef.setBeta(-1, r, s, max_beta, beta_start);
			coef.setConstB(totalVar-1,-3/2);
			constraintMat.push_back(coef);
		}
		int len=min(betaSr.end-betaSr.start+1, betaSs.end-betaSs.start+1)-1;
		Constraint coef; //coef.resize(totalVar, 0);
		for (int i=betaSr.start; i<=betaSr.end; i++)
		for (int j=betaSs.start; j<=betaSs.end; j++){
			coef.setDist(-1/8, i, j, seq_len);
		}
		coef.setBeta(len, r, s, max_beta, beta_start);
		coef.setConstB(totalVar-1,(betaSr.end-betaSr.start+1)*(betaSs.end-betaSs.start+1)*3/2);
		constraintMat.push_back(coef);
	}
	cerr << seq_len << " " << totalVar << "-" << beta_start << " " << constraintMat.size() << endl;
	// number of nonlocal contacts a strand residue s_i can have with other residues <10
	for (int i=0; i<seq_len; i++) {
		if (ss_idx[i]<0 || ss_idx[i]>=100)
			continue;
		Constraint coef; //coef.resize(totalVar, 0);
		for (int j=0; j<seq_len; j++){
			if (abs(i-j)<=6)
				continue;
			coef.setDist(-1/8, i, j, seq_len);
		}
		coef.setConstB(totalVar-1,10-(seq_len-12)*3/2);
		constraintMat.push_back(coef);
	}
	// If two nonlocal residues i and j make a close contact, then at least one of the neighboring residues of strand residue i and j should make a contact.
	cerr << seq_len << " " << totalVar << "-" << beta_start << " " << constraintMat.size() << endl;
	for (int i=1; i<seq_len-3; i++) 
	for (int j=i+6; j<seq_len-2; j++){
		if (ss_idx[i]<0 || ss_idx[i]>=100 || ss_idx[j]<0 || ss_idx[j]>=100)
			continue;
		Constraint coef; //coef.resize(totalVar, 0);
		coef.setDist(1/8, i-1, j, seq_len);
		coef.setDist(1/8, i+1, j, seq_len);
		coef.setDist(-1/8, i, j, seq_len);
		coef.setConstB(totalVar-1,3/2);
		constraintMat.push_back(coef);

		Constraint coef1; //coef1.resize(totalVar, 0);
		coef1.setDist(1/8, i, j-1, seq_len);
		coef1.setDist(1/8, i, j+1, seq_len);
		coef1.setDist(-1/8, i, j, seq_len);
		coef1.setConstB(totalVar-1,3/2);
		constraintMat.push_back(coef1);
	}
	cerr << seq_len << " " << totalVar << "-" << beta_start << " " << constraintMat.size() << endl;

	// If residue i, i+2 contact residue j, j+2 specifically, then residue i+1 should also contact residue j+1 
	for (int i=0; i<seq_len-9; i++) 
	for (int j=i+6; j<seq_len-2; j++){
		if (ss_idx[i]<0 || ss_idx[i]>=100 || ss_idx[j]<0 || ss_idx[j]>=100)
			continue;
		Constraint coef; //coef.resize(totalVar, 0);
		coef.setDist(-1/2, i, j, seq_len);
		coef.setDist(-1/2, i+2, j+2, seq_len);
		coef.setDist(1, i+1, j+1, seq_len);
		coef.setConstB(totalVar-1,2);
		constraintMat.push_back(coef);
	}

	cerr << seq_len << " " << totalVar << "-" << beta_start << " " << constraintMat.size() << endl;
	// If residue i, i+2 contact residue j+2, j specifically, then residue i+1 should also contact residue j+1 
	for (int i=0; i<seq_len-9; i++) 
	for (int j=i+6; j<seq_len-2; j++){
		if (ss_idx[i]<0 || ss_idx[i]>=100 || ss_idx[j]<0 || ss_idx[j]>=100)
			continue;
		Constraint coef; //coef.resize(totalVar, 0);
		coef.setDist(-1/2, i, j+2, seq_len);
		coef.setDist(-1/2, i+2, j, seq_len);
		coef.setDist(1, i+1, j+1, seq_len);
		coef.setConstB(totalVar-1,2);
		constraintMat.push_back(coef);
	}
	print(seq_len, totalVar-beta_start, alpha_idx-99, beta_idx+1);
}

void Constraints::print(int seq_len, int assistVarNum, int num_alpha, int num_beta)
{
	cout << seq_len << " " << assistVarNum << " " << constraintMat.size() << " " << num_alpha << " " << num_beta <<endl;
	for (int i=0; i<(int)constraintMat.size(); i++){
		Constraint cs=constraintMat[i];
		for (int j=0; j<(int)cs.co.size(); j++)
			cout << i+1 << " " << cs.co[j].idx+1 << " " << cs.co[j].val << endl;
	}
}

