#include <cstdlib>
#include <iostream>
#include <fstream>
#include <time.h>
#include <stdlib.h>
#include <values.h>
#include <math.h>

#include <BALL/KERNEL/protein.h>
#include <BALL/KERNEL/atomContainer.h>
#include <BALL/FORMAT/PDBFile.h>
#include <BALL/KERNEL/residue.h>
#include <BALL/KERNEL/atom.h>

#include "ScoringFunction.h"
#include "residue.h"

using namespace std;
using namespace BALL;

#define SCORE_EPAD		1
#define SCORE_EPADLocal	2	

#define DEF_OUT_DIR	"./"

typedef vector<RESIDUE> BACKBONE;
typedef vector<string> STR_VEC;

vector<pair<ScoringFunction*, int> > mScoringFunctionList;
string feature_f, feature_f1;

string arrResidueName[21]={"ALA","ARG","ASN","ASP","CYS","GLN","GLU","GLY","HIS","ILE", "LEU","LYS","MET","PHE","PRO","SER","THR","TRP","TYR","VAL","HET"};

// read the pdb file to get the structure of a protein and fill into BACKBONE vector
// return -1 if not succeed
int GetConformation(string pdb_file, vector<RESIDUE> &g_backbone)
{
	g_backbone.clear();
	Protein* protein = new Protein;
	// read the pdb protein file
	PDBFile* pdb=0;
	try{
		pdb = new PDBFile(pdb_file, std::ios::in);
	}catch(Exception::FileNotFound){
		cerr<<"Can not find protein file "<<pdb_file <<endl;
		return -1;
	}

	*pdb >> *protein;
	cerr << "protein data read with " << protein->countChains() << " chains; ";
	cerr << protein->countResidues() << " residues; " << protein->countPDBAtoms() << " atoms\n";

	string start_res_str=protein->beginResidue()->getID();
	int start_res=atoi(start_res_str.c_str());
	for (int i=0; i<start_res-1;i++){
		RESIDUE res;
		g_backbone.push_back(res);
	}

	int nRes=0, lastRes=start_res;
	// retrieve the residue and atoms from the Protein structure and insert into the bbq acceptable structure
	for(ResidueIterator resIter=protein->beginResidue(); resIter!=protein->endResidue(); resIter++){
		RESIDUE res;
		res.name= resIter->getFullName(Residue::NO_VARIANT_EXTENSIONS);
		for (int i=0; i<20; i++)
			if (arrResidueName[i]==res.name){
				res.name_idx=i; break;
			}
		res.seq = nRes; //cerr << resIter->getID() << endl;
		if (res.name.size()<3)
			break; //continue;

		string res_str=resIter->getID();
		int resNo=atoi(res_str.c_str());
		//cerr << "last=" << lastRes << "; this=" << resNo << ":" << res_str << endl;
		for (int i=0; i<resNo-lastRes-1;i++){
			RESIDUE res;
			g_backbone.push_back(res);
		}
		lastRes=resNo;

		Vector3 pos;
		for(AtomIterator at=resIter->beginAtom();at!=resIter->endAtom();at++){
			pos = at->getPosition();
			ATOM atom(at->getName(), res.seq);

			atom.x = pos.x;
			atom.y = pos.y;
			atom.z = pos.z;

			if (atom.name!="N" && atom.name!="O" && atom.name!="CA" && atom.name!="CB" && atom.name!="C" && atom.name!="NH" && atom.name!="HN")
				continue;
			res.addAtom( atom);
		}
		nRes++;
		
		if (res.getAtom("CA").name=="")
			break;

		ATOM C("C", res.seq), CB("CB", res.seq), N("O", res.seq), O("N", res.seq);
		if (res.getAtom("CB").name==""&&res.name!="GLY")
			res.addAtom(CB);
		if (res.getAtom("C").name=="")
			res.addAtom(C);
		if (res.getAtom("O").name=="")
			res.addAtom(O);
		if (res.getAtom("N").name=="")
			res.addAtom(N);

		g_backbone.push_back(res);
		//cerr << "push_back: " << res.name << "  " << g_backbone.size() << endl;
	}
	
	if (protein) delete protein;
	if (pdb) delete pdb;

	cerr << "getConformation done." << endl;
	return start_res;
}

void Usage(){
	cerr<<"Usage v2.3: testEPAD [-d dir] [-A feature_file] [-L list_file]"<<endl;
	cerr<<"\t-d dir: the directory that contains the protein files. Default is the current directory." << endl;
	cerr<<"\t-L list_file: this file lists the decoy files in the dir defined in the -d option." << endl;
	cerr<<"\t-A feature_file: it is the *.epad file generated for EPAD." << endl;
	cerr<<"Examples: " << endl;
	cerr<<"\t ./testEPAD -A ~/SEQ/1enhA.epad -L list_1enhA 2>err 1>1enhA.res" << endl;
}

// Initialize the potential functions
void setScoringFunction(int scoring)
{
	int startRes, endRes;
	ScoringFunction* scoringFunction=0;
	switch (scoring)
	{
	case SCORE_EPADLocal:
		scoringFunction=new ScoringByEPADLocal();
		if (feature_f1 !="")
			scoringFunction->Init("./config/modelCNFo1SS.dat", feature_f1, "./config/cnf_label_SS_1.dat");
		else
			scoringFunction->Init("./config/modelCNFo1SS.dat", feature_f, "./config/cnf_label_SS_1.dat");
		mScoringFunctionList.push_back(pair<ScoringFunction*, int>(scoringFunction, scoring));
		break;
	case SCORE_EPAD:
		scoringFunction=new ScoringByEPAD();
		scoringFunction->Init("./config/epad.cfg", feature_f, "");

		mScoringFunctionList.push_back(pair<ScoringFunction*, int>(scoringFunction, scoring));
		break;
	}
}

int main(int argc, char** argv)
{
	if(argc<2) {
		Usage();
		exit(-1);
	}
	
	string dir = DEF_OUT_DIR; // default input directory
	extern char* optarg;
	extern int optind;
	char c=0;
	string file_list="";
	feature_f="", feature_f1="";
	while((c=getopt(argc,argv,":A:d:L:"))!=EOF){
		cerr << "<" <<c<<">";
		switch(c)
		{
		case 'A': // feature file
			feature_f=optarg;
			feature_f1=feature_f+"_local";
			break;
		case 'd': // directory
			dir = optarg;
			cerr << "dir=" << dir << endl;
			break;
		case 'L':
			file_list = optarg;
			break;
		default:
			Usage();
			exit(-1);
		}
	}

	cerr << file_list << endl;

	// Initialize EPAD and EPADLocal
	setScoringFunction(SCORE_EPAD);
	setScoringFunction(SCORE_EPADLocal);

	// get the list of decoys
	STR_VEC seqs;
	ifstream idf(file_list.c_str());
	string line;
	while (getline(idf, line)){
		//cerr << "line: " << line << "\n";
		seqs.push_back(line);
	}
	idf.close();
	
	// scan the decoy list and calculate the potential of each decoy.
	BACKBONE g_backbone;
	for (int n=0; n<(int)seqs.size(); n++) {
		cerr << "Protein file: " << seqs[n] << endl;

		// read the decoy file to retrieve the structure
		string fname = dir + "/" + seqs[n];
		// We need the decoy structure match the original sequence. However, there are cases
		// that the decoy has missing residues. startRes records the sequence number of the first
		// non-missing residue.
		int startRes=GetConformation( fname, g_backbone);
		if (g_backbone.size()==0) {
			cerr << "Empty Protein file: " << fname << endl;
			continue;
		}

		// calculate the potentials
		cout << seqs[n] << " ";
		double total_energy=0;
		for (int i=0; i<(int)mScoringFunctionList.size(); i++) {
			ScoringFunction* sf = mScoringFunctionList[i].first;
			sf->startRes=startRes;
			double energy = sf->evaluate(g_backbone, 0, -1);
			total_energy+=energy;
			cout << i <<"(" << sf->mName << "):" << energy << " ";
		}
		cout << "Energy= " << total_energy << endl;
	}
}
